/*
 * Connect a number of files together and put a call to each in a list
 * at the start, ending RTS. [68000] 
 *
 * No size limitations, apart from memory.
 *
 * (c) 1991 Richard Kettlewell.
 *
 * Permission is given to distribute this program provided that this
 * comment is left intact and that there is no commercial interest.
 *
 * 0.1		SuperBASIC version, maximum 32k per file
 * 0.2		First C version, any size file
 * 0.21		Error handling improved slightly
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <qdos.h>

char _PROG_NAME[] = "ProcLink 0.21";	/* 0.1 was SuperBASIC version */
long _MNEED = 200000;					/* More than enough */
#define MAXFILES 16						/* Enough */
#define COPYSIZE 150000					/* Much more than enough */
struct filedef {
	int filelength;						/* File length for creation of call table */
	FILE *fp;
};

main(argc,argv)
int argc;
char **argv;
{
	struct filedef filelist[MAXFILES];	/* List of file information */
	char buffer[80];					/* Input buffer */
	int i, files;						/* Various file counts */
	FILE *fp;							/* Various files */
	long position0, position1,			/* Positions within file */
		offset;							/* Offset */
	void *copyarea;						/* Buffer for copy loop */
	unsigned bytes;						/* Bytes copied in last gulp */
	char flag;							/* A flag */

	flag = (fgetchid(stdin) != fgetchid(stdout));
	/*
	 * Program information if any parameters
	 */
	if(argc != 1) {
		printf("%s (c) 1991 Richard Kettlewell\n",_PROG_NAME);
		puts("Either input directly from keyboard, or redirect stdin with\n< operator e.g. <stdprocs_dat\nInput is list of files, with a blank line at the end and\nthen the output file on its own.\nIncrease memory of this program [if necessary] with % e.g. %400000");
		printf("[Default is %ld]\n", _MNEED);
		return -15;
	}
	files = 0;
	printf("%s (c) 1991 Richard Kettlewell\n\nFor help, call with parameters\nType 'abort' to escape\n\n",_PROG_NAME);

	/*
	 * Allocate memory
	 */
	if(!(copyarea = malloc(COPYSIZE))) {
		puts("Out of memory");
		return -3;
	}
	/*
	 * Get first line
	 */
	printf("Input filename> ");
	if(!gets(buffer))
		return -21;
	if(flag)
		puts(buffer);
	while(*buffer) {
		if(!stricmp(buffer,"abort"))
			return -1;
		/*
		 * Check file table has some space left
		 */
		if(files > MAXFILES) {
			printf("Can only manage %d files\n",MAXFILES);
			return -4;
		}
		if(!(fp = filelist[files].fp = fopen(buffer, "r")))
			printf("Failed to open input file %s\n",buffer);
		else {
			/*
			 * Find file length
			 * fseek() returns 0 for success, non0 for pipes &c
			 */
			if(fseek(fp, 0L, SEEK_END)) {
				printf("fseek() fails on file %s\n");
				return -19;
			}
			printf("File %s has length %ld\n", buffer, filelist[files].filelength = ftell(fp));
			/*
			 * Return to start of file
			 */
			fseek(fp, 0L, SEEK_SET);
			files++;
		}
		/*
		 * Get next filename
		 */
		printf("Input filename> ");
		if(!gets(buffer))
			return -21;
		if(flag)
			puts(buffer);
	}
	/*
	 * Get output filename
	 */
	printf("Output filename> ");
	if(!gets(buffer))
		return -21;
	if(flag)
		puts(buffer);
	if(!stricmp("abort",buffer))
		return -1;
	if(!(fp = fopen(buffer, "w"))) {
		printf("Failed to open output file %s\n",buffer);
		return -6;
	}
	position0 = 0L;
	position1 = 10 * files + 2;

	/*
	 * Write calling sequence
	 */
	for(i = 0; i < files; i++) {
		putc('\x22', fp);				/* MOVE.L #p1-(*+2),D1 */
		putc('\x3C', fp);
		offset = position1 - position0;
		fwrite(&offset, sizeof(offset), 1, fp);
		putc('\x4E', fp);				/* JSR -8(PC,D1.L) */
		putc('\xBB', fp);
		putc('\x18', fp);
		putc('\xF8', fp);
		position0 += 10;
		position1 += filelist[i].filelength;
		position1 += position1 & 1;
	}
	putc('\x4E', fp);					/* RTS */
	putc('\x75', fp);
	printf("Size of header = %ld bytes\n",position0 += 2);

	/*
	 * Output files
	 */
	for(i = 0; i < files; i++) {
		/*
		 * Copy one file
		 */
		do {
			bytes = fread(copyarea, 1, COPYSIZE, filelist[i].fp);
			if(bytes) {
				fwrite(copyarea, 1, bytes, fp);
				position0 += bytes;
			}
		} while (bytes);
		/*
		 * If file length is odd, round up to even
		 */
		if(position0 & 1) {
			putc('\x00', fp);
			position0++;
		}
		/*
		 * Finished with that file
		 */
		fclose(filelist[i].fp);
	}
	printf("Size of output file %s is %ld", buffer, position0);
	/*
	 * Tidy exit
	 */
	fclose(fp);
	free(copyarea);
	return 0;
}
